#ifndef AV_VERSION_INT
# define AV_VERSION_INT(a,b,c) 0x7FFFFFFF
#endif
const ff_codec_t ff_audio_lookup[] = {
  { BUF_AUDIO_WMAV1, AV_CODEC_ID_WMAV1, "MS Windows Media Audio 1 (ffmpeg)" },
  { BUF_AUDIO_WMAV2, AV_CODEC_ID_WMAV2, "MS Windows Media Audio 2 (ffmpeg)" },
  { BUF_AUDIO_WMAPRO, AV_CODEC_ID_WMAPRO, "MS Windows Media Audio Professional (ffmpeg)" },
  { BUF_AUDIO_14_4, AV_CODEC_ID_RA_144, "Real 14.4 (ffmpeg)" },
  { BUF_AUDIO_28_8, AV_CODEC_ID_RA_288, "Real 28.8 (ffmpeg)" },
  { BUF_AUDIO_MPEG, AV_CODEC_ID_MP3, "MP3 (ffmpeg)" },
  { BUF_AUDIO_MP3ADU, AV_CODEC_ID_MP3ADU, "MPEG-3 adu (ffmpeg)" },
  { BUF_AUDIO_MSADPCM, AV_CODEC_ID_ADPCM_MS, "MS ADPCM (ffmpeg)" },
  { BUF_AUDIO_QTIMAADPCM, AV_CODEC_ID_ADPCM_IMA_QT, "QT IMA ADPCM (ffmpeg)" },
  { BUF_AUDIO_MSIMAADPCM, AV_CODEC_ID_ADPCM_IMA_WAV, "MS IMA ADPCM (ffmpeg)" },
  { BUF_AUDIO_DK3ADPCM, AV_CODEC_ID_ADPCM_IMA_DK3, "Duck DK3 ADPCM (ffmpeg)" },
  { BUF_AUDIO_DK4ADPCM, AV_CODEC_ID_ADPCM_IMA_DK4, "Duck DK4 ADPCM (ffmpeg)" },
  { BUF_AUDIO_VQA_IMA, AV_CODEC_ID_ADPCM_IMA_WS, "Westwood Studios IMA (ffmpeg)" },
  { BUF_AUDIO_SMJPEG_IMA, AV_CODEC_ID_ADPCM_IMA_SMJPEG, "SMJPEG IMA (ffmpeg)" },
  { BUF_AUDIO_XA_ADPCM, AV_CODEC_ID_ADPCM_XA, "CD-ROM/XA ADPCM (ffmpeg)" },
  { BUF_AUDIO_4X_ADPCM, AV_CODEC_ID_ADPCM_4XM, "4X ADPCM (ffmpeg)" },
  { BUF_AUDIO_EA_ADPCM, AV_CODEC_ID_ADPCM_EA, "Electronic Arts ADPCM (ffmpeg)" },
  { BUF_AUDIO_MULAW, AV_CODEC_ID_PCM_MULAW, "mu-law logarithmic PCM (ffmpeg)" },
  { BUF_AUDIO_ALAW, AV_CODEC_ID_PCM_ALAW, "A-law logarithmic PCM (ffmpeg)" },
  { BUF_AUDIO_ROQ, AV_CODEC_ID_ROQ_DPCM, "RoQ DPCM (ffmpeg)" },
  { BUF_AUDIO_INTERPLAY, AV_CODEC_ID_INTERPLAY_DPCM, "Interplay DPCM (ffmpeg)" },
  { BUF_AUDIO_MAC3, AV_CODEC_ID_MACE3, "MACE 3:1 (ffmpeg)" },
  { BUF_AUDIO_MAC6, AV_CODEC_ID_MACE6, "MACE 6:1 (ffmpeg)" },
  { BUF_AUDIO_XAN_DPCM, AV_CODEC_ID_XAN_DPCM, "Origin Xan DPCM (ffmpeg)" },
  { BUF_AUDIO_VMD, AV_CODEC_ID_VMDAUDIO, "Sierra VMD Audio (ffmpeg)" },
  { BUF_AUDIO_FLAC, AV_CODEC_ID_FLAC, "FLAC (ffmpeg)" },
  { BUF_AUDIO_SHORTEN, AV_CODEC_ID_SHORTEN, "Shorten (ffmpeg)" },
  { BUF_AUDIO_ALAC, AV_CODEC_ID_ALAC, "ALAC (ffmpeg)" },
  { BUF_AUDIO_QDESIGN2, AV_CODEC_ID_QDM2, "QDesign (ffmpeg)" },
  { BUF_AUDIO_COOK, AV_CODEC_ID_COOK, "RealAudio Cooker (ffmpeg)" },
  { BUF_AUDIO_ATRK, AV_CODEC_ID_ATRAC3, "ATRAC 3 (ffmpeg)" },
  { BUF_AUDIO_TRUESPEECH, AV_CODEC_ID_TRUESPEECH, "TrueSpeech (ffmpeg)" },
  { BUF_AUDIO_TTA, AV_CODEC_ID_TTA, "True Audio Lossless (ffmpeg)" },
  { BUF_AUDIO_SMACKER, AV_CODEC_ID_SMACKAUDIO, "Smacker (ffmpeg)" },
  { BUF_AUDIO_FLVADPCM, AV_CODEC_ID_ADPCM_SWF, "Flash ADPCM (ffmpeg)" },
  { BUF_AUDIO_WAVPACK, AV_CODEC_ID_WAVPACK, "WavPack (ffmpeg)" },
  { BUF_AUDIO_AMR_NB, AV_CODEC_ID_AMR_NB, "AMR narrow band (ffmpeg)" },
  { BUF_AUDIO_AMR_WB, AV_CODEC_ID_AMR_WB, "AMR wide band (ffmpeg)" },
  { BUF_AUDIO_A52, AV_CODEC_ID_AC3, "AC-3 (ffmpeg)" },
  { BUF_AUDIO_EAC3, AV_CODEC_ID_EAC3, "E-AC-3 (ffmpeg)" },
  { BUF_AUDIO_AAC, AV_CODEC_ID_AAC, "MPEG4 (ffmpeg)" },
  { BUF_AUDIO_AAC_LATM, AV_CODEC_ID_AAC_LATM, "AAC LATM (ffmpeg)" },
  { BUF_AUDIO_ADPCM_G726, AV_CODEC_ID_ADPCM_G726, "ADPCM G726 (ffmpeg)" },
  { BUF_AUDIO_QCLP, AV_CODEC_ID_QCELP, "QualComm Purevoice (ffmpeg)" },
  { BUF_AUDIO_OPUS, AV_CODEC_ID_OPUS, "Opus Audio (ffmpeg)" },
  { BUF_AUDIO_DTS, AV_CODEC_ID_DTS, "DTS (ffmpeg)" },
  { BUF_AUDIO_TRUEHD, AV_CODEC_ID_TRUEHD, "TrueHD (ffmpeg)" },
};

static const uint32_t supported_audio_types[] = {
  BUF_AUDIO_WMAV1,
  BUF_AUDIO_WMAV2,
  BUF_AUDIO_WMAPRO,
  BUF_AUDIO_14_4,
  BUF_AUDIO_28_8,
  BUF_AUDIO_MPEG,
  BUF_AUDIO_MP3ADU,
  BUF_AUDIO_MSADPCM,
  BUF_AUDIO_QTIMAADPCM,
  BUF_AUDIO_MSIMAADPCM,
  BUF_AUDIO_DK3ADPCM,
  BUF_AUDIO_DK4ADPCM,
  BUF_AUDIO_VQA_IMA,
  BUF_AUDIO_SMJPEG_IMA,
  BUF_AUDIO_XA_ADPCM,
  BUF_AUDIO_4X_ADPCM,
  BUF_AUDIO_EA_ADPCM,
  BUF_AUDIO_MULAW,
  BUF_AUDIO_ALAW,
  BUF_AUDIO_ROQ,
  BUF_AUDIO_INTERPLAY,
  BUF_AUDIO_MAC3,
  BUF_AUDIO_MAC6,
  BUF_AUDIO_XAN_DPCM,
  BUF_AUDIO_VMD,
  BUF_AUDIO_FLAC,
  BUF_AUDIO_SHORTEN,
  BUF_AUDIO_ALAC,
  BUF_AUDIO_QDESIGN2,
  BUF_AUDIO_COOK,
  BUF_AUDIO_ATRK,
  BUF_AUDIO_TRUESPEECH,
  BUF_AUDIO_TTA,
  BUF_AUDIO_SMACKER,
  BUF_AUDIO_FLVADPCM,
  BUF_AUDIO_WAVPACK,
  BUF_AUDIO_AMR_NB,
  BUF_AUDIO_AMR_WB,
  BUF_AUDIO_A52,
  BUF_AUDIO_EAC3,
  BUF_AUDIO_AAC,
  BUF_AUDIO_AAC_LATM,
  BUF_AUDIO_ADPCM_G726,
  BUF_AUDIO_QCLP,
  BUF_AUDIO_OPUS,
  BUF_AUDIO_DTS,
  BUF_AUDIO_TRUEHD,
  0,
};
